"""
Basically a contains parser that is faster, because it tries to parse only
parts and if anything changes, it only reparses the changed parts.

It works with a simple diff in the beginning and will try to reuse old parser
fragments.
"""
import re
import difflib
from collections import namedtuple
import logging

from rp.libs.parso.utils import split_lines
from rp.libs.parso.python.parser import Parser
from rp.libs.parso.python.tree import EndMarker
from rp.libs.parso.python.tokenize import PythonToken
from rp.libs.parso.python.token import PythonTokenTypes

LOG = logging.getLogger(__name__)
DEBUG_DIFF_PARSER = False

_INDENTATION_TOKENS = 'INDENT', 'ERROR_DEDENT', 'DEDENT'


def _get_previous_leaf_if_indentation(leaf):
    while leaf and leaf.type == 'error_leaf' \
            and leaf.token_type in _INDENTATION_TOKENS:
        leaf = leaf.get_previous_leaf()
    return leaf


def _get_next_leaf_if_indentation(leaf):
    while leaf and leaf.type == 'error_leaf' \
            and leaf.token_type in _INDENTATION_TOKENS:
        leaf = leaf.get_previous_leaf()
    return leaf


def _assert_valid_graph(node):
    """
    Checks if the parent/children relationship is correct.

    This is a check that only runs during debugging/testing.
    """
    try:
        children = node.children
    except AttributeError:
        # Ignore INDENT is necessary, because indent/dedent tokens don't
        # contain value/prefix and are just around, because of the tokenizer.
        if node.type == 'error_leaf' and node.token_type in _INDENTATION_TOKENS:
            assert not node.value
            assert not node.prefix
            return

        # Calculate the content between two start positions.
        previous_leaf = _get_previous_leaf_if_indentation(node.get_previous_leaf())
        if previous_leaf is None:
            content = node.prefix
            previous_start_pos = 1, 0
        else:
            assert previous_leaf.end_pos <= node.start_pos, \
                (previous_leaf, node)

            content = previous_leaf.value + node.prefix
            previous_start_pos = previous_leaf.start_pos

        if '\n' in content or '\r' in content:
            splitted = split_lines(content)
            line = previous_start_pos[0] + len(splitted) - 1
            actual = line, len(splitted[-1])
        else:
            actual = previous_start_pos[0], previous_start_pos[1] + len(content)

        assert node.start_pos == actual, (node.start_pos, actual)
    else:
        for child in children:
            assert child.parent == node, (node, child)
            _assert_valid_graph(child)


def _get_debug_error_message(module, old_lines, new_lines):
    current_lines = split_lines(module.get_code(), keepends=True)
    current_diff = difflib.unified_diff(new_lines, current_lines)
    old_new_diff = difflib.unified_diff(old_lines, new_lines)
    import rp.libs.parso as parso
    return (
        "There's an issue with the diff parser. Please "
        "report (parso v%s) - Old/New:\n%s\nActual Diff (May be empty):\n%s"
        % (parso.__version__, ''.join(old_new_diff), ''.join(current_diff))
    )


def _get_last_line(node_or_leaf):
    last_leaf = node_or_leaf.get_last_leaf()
    if _ends_with_newline(last_leaf):
        return last_leaf.start_pos[0]
    else:
        return last_leaf.end_pos[0]


def _skip_dedent_error_leaves(leaf):
    while leaf is not None and leaf.type == 'error_leaf' and leaf.token_type == 'DEDENT':
        leaf = leaf.get_previous_leaf()
    return leaf


def _ends_with_newline(leaf, suffix=''):
    leaf = _skip_dedent_error_leaves(leaf)

    if leaf.type == 'error_leaf':
        typ = leaf.token_type.lower()
    else:
        typ = leaf.type

    return typ == 'newline' or suffix.endswith('\n') or suffix.endswith('\r')


def _flows_finished(pgen_grammar, stack):
    """
    if, while, for and try might not be finished, because another part might
    still be parsed.
    """
    for stack_node in stack:
        if stack_node.nonterminal in ('if_stmt', 'while_stmt', 'for_stmt', 'try_stmt'):
            return False
    return True


def _func_or_class_has_suite(node):
    if node.type == 'decorated':
        node = node.children[-1]
    if node.type in ('async_funcdef', 'async_stmt'):
        node = node.children[-1]
    return node.type in ('classdef', 'funcdef') and node.children[-1].type == 'suite'


def _suite_or_file_input_is_valid(pgen_grammar, stack):
    if not _flows_finished(pgen_grammar, stack):
        return False

    for stack_node in reversed(stack):
        if stack_node.nonterminal == 'decorator':
            # A decorator is only valid with the upcoming function.
            return False

        if stack_node.nonterminal == 'suite':
            # If only newline is in the suite, the suite is not valid, yet.
            return len(stack_node.nodes) > 1
    # Not reaching a suite means that we're dealing with file_input levels
    # where there's no need for a valid statement in it. It can also be empty.
    return True


def _is_flow_node(node):
    if node.type == 'async_stmt':
        node = node.children[1]
    try:
        value = node.children[0].value
    except AttributeError:
        return False
    return value in ('if', 'for', 'while', 'try', 'with')


class _PositionUpdatingFinished(Exception):
    pass


def _update_positions(nodes, line_offset, last_leaf):
    for node in nodes:
        try:
            children = node.children
        except AttributeError:
            # Is a leaf
            node.line += line_offset
            if node is last_leaf:
                raise _PositionUpdatingFinished
        else:
            _update_positions(children, line_offset, last_leaf)


class DiffParser(object):
    """
    An advanced form of parsing a file faster. Unfortunately comes with huge
    side effects. It changes the given module.
    """
    def __init__(self, pgen_grammar, tokenizer, module):
        self._pgen_grammar = pgen_grammar
        self._tokenizer = tokenizer
        self._module = module

    def _reset(self):
        self._copy_count = 0
        self._parser_count = 0

        self._nodes_tree = _NodesTree(self._module)

    def update(self, old_lines, new_lines):
        '''
        The algorithm works as follows:

        Equal:
            - Assure that the start is a newline, otherwise parse until we get
              one.
            - Copy from parsed_until_line + 1 to max(i2 + 1)
            - Make sure that the indentation is correct (e.g. add DEDENT)
            - Add old and change positions
        Insert:
            - Parse from parsed_until_line + 1 to min(j2 + 1), hopefully not
              much more.

        Returns the new module node.
        '''
        LOG.debug('diff parser start')
        # Reset the used names cache so they get regenerated.
        self._module._used_names = None

        self._parser_lines_new = new_lines

        self._reset()

        line_length = len(new_lines)
        sm = difflib.SequenceMatcher(None, old_lines, self._parser_lines_new)
        opcodes = sm.get_opcodes()
        LOG.debug('line_lengths old: %s; new: %s' % (len(old_lines), line_length))

        for operation, i1, i2, j1, j2 in opcodes:
            LOG.debug('-> code[%s] old[%s:%s] new[%s:%s]',
                      operation, i1 + 1, i2, j1 + 1, j2)

            if j2 == line_length and new_lines[-1] == '':
                # The empty part after the last newline is not relevant.
                j2 -= 1

            if operation == 'equal':
                line_offset = j1 - i1
                self._copy_from_old_parser(line_offset, i2, j2)
            elif operation == 'replace':
                self._parse(until_line=j2)
            elif operation == 'insert':
                self._parse(until_line=j2)
            else:
                assert operation == 'delete'

        # With this action all change will finally be applied and we have a
        # changed module.
        self._nodes_tree.close()

        if DEBUG_DIFF_PARSER:
            # If there is reasonable suspicion that the diff parser is not
            # behaving well, this should be enabled.
            try:
                assert self._module.get_code() == ''.join(new_lines)
                _assert_valid_graph(self._module)
            except AssertionError:
                print(_get_debug_error_message(self._module, old_lines, new_lines))
                raise

        last_pos = self._module.end_pos[0]
        if last_pos != line_length:
            raise Exception(
                ('(%s != %s) ' % (last_pos, line_length))
                + _get_debug_error_message(self._module, old_lines, new_lines)
            )
        LOG.debug('diff parser end')
        return self._module

    def _enabled_debugging(self, old_lines, lines_new):
        if self._module.get_code() != ''.join(lines_new):
            LOG.warning('parser issue:\n%s\n%s', ''.join(old_lines), ''.join(lines_new))

    def _copy_from_old_parser(self, line_offset, until_line_old, until_line_new):
        last_until_line = -1
        while until_line_new > self._nodes_tree.parsed_until_line:
            parsed_until_line_old = self._nodes_tree.parsed_until_line - line_offset
            line_stmt = self._get_old_line_stmt(parsed_until_line_old + 1)
            if line_stmt is None:
                # Parse 1 line at least. We don't need more, because we just
                # want to get into a state where the old parser has statements
                # again that can be copied (e.g. not lines within parentheses).
                self._parse(self._nodes_tree.parsed_until_line + 1)
            else:
                p_children = line_stmt.parent.children
                index = p_children.index(line_stmt)

                from_ = self._nodes_tree.parsed_until_line + 1
                copied_nodes = self._nodes_tree.copy_nodes(
                    p_children[index:],
                    until_line_old,
                    line_offset
                )
                # Match all the nodes that are in the wanted range.
                if copied_nodes:
                    self._copy_count += 1

                    to = self._nodes_tree.parsed_until_line

                    LOG.debug('copy old[%s:%s] new[%s:%s]',
                              copied_nodes[0].start_pos[0],
                              copied_nodes[-1].end_pos[0] - 1, from_, to)
                else:
                    # We have copied as much as possible (but definitely not too
                    # much). Therefore we just parse a bit more.
                    self._parse(self._nodes_tree.parsed_until_line + 1)
            # Since there are potential bugs that might loop here endlessly, we
            # just stop here.
            assert last_until_line != self._nodes_tree.parsed_until_line, last_until_line
            last_until_line = self._nodes_tree.parsed_until_line

    def _get_old_line_stmt(self, old_line):
        leaf = self._module.get_leaf_for_position((old_line, 0), include_prefixes=True)

        if _ends_with_newline(leaf):
            leaf = leaf.get_next_leaf()
        if leaf.get_start_pos_of_prefix()[0] == old_line:
            node = leaf
            while node.parent.type not in ('file_input', 'suite'):
                node = node.parent

            # Make sure that if only the `else:` line of an if statement is
            # copied that not the whole thing is going to be copied.
            if node.start_pos[0] >= old_line:
                return node
        # Must be on the same line. Otherwise we need to parse that bit.
        return None

    def _parse(self, until_line):
        """
        Parses at least until the given line, but might just parse more until a
        valid state is reached.
        """
        last_until_line = 0
        while until_line > self._nodes_tree.parsed_until_line:
            node = self._try_parse_part(until_line)
            nodes = node.children

            self._nodes_tree.add_parsed_nodes(nodes)
            LOG.debug(
                'parse_part from %s to %s (to %s in part parser)',
                nodes[0].get_start_pos_of_prefix()[0],
                self._nodes_tree.parsed_until_line,
                node.end_pos[0] - 1
            )
            # Since the tokenizer sometimes has bugs, we cannot be sure that
            # this loop terminates. Therefore assert that there's always a
            # change.
            assert last_until_line != self._nodes_tree.parsed_until_line, last_until_line
            last_until_line = self._nodes_tree.parsed_until_line

    def _try_parse_part(self, until_line):
        """
        Sets up a normal parser that uses a spezialized tokenizer to only parse
        until a certain position (or a bit longer if the statement hasn't
        ended.
        """
        self._parser_count += 1
        # TODO speed up, shouldn't copy the whole list all the time.
        # memoryview?
        parsed_until_line = self._nodes_tree.parsed_until_line
        lines_after = self._parser_lines_new[parsed_until_line:]
        tokens = self._diff_tokenize(
            lines_after,
            until_line,
            line_offset=parsed_until_line
        )
        self._active_parser = Parser(
            self._pgen_grammar,
            error_recovery=True
        )
        return self._active_parser.parse(tokens=tokens)

    def _diff_tokenize(self, lines, until_line, line_offset=0):
        is_first_token = True
        omitted_first_indent = False
        indents = []
        tokens = self._tokenizer(lines, (1, 0))
        stack = self._active_parser.stack
        for typ, string, start_pos, prefix in tokens:
            start_pos = start_pos[0] + line_offset, start_pos[1]
            if typ == PythonTokenTypes.INDENT:
                indents.append(start_pos[1])
                if is_first_token:
                    omitted_first_indent = True
                    # We want to get rid of indents that are only here because
                    # we only parse part of the file. These indents would only
                    # get parsed as error leafs, which doesn't make any sense.
                    is_first_token = False
                    continue
            is_first_token = False

            # In case of omitted_first_indent, it might not be dedented fully.
            # However this is a sign for us that a dedent happened.
            if typ == PythonTokenTypes.DEDENT \
                    or typ == PythonTokenTypes.ERROR_DEDENT \
                    and omitted_first_indent and len(indents) == 1:
                indents.pop()
                if omitted_first_indent and not indents:
                    # We are done here, only thing that can come now is an
                    # endmarker or another dedented code block.
                    typ, string, start_pos, prefix = next(tokens)
                    if '\n' in prefix or '\r' in prefix:
                        prefix = re.sub(r'[^\n\r]+\Z', '', prefix)
                    else:
                        assert start_pos[1] >= len(prefix), repr(prefix)
                        if start_pos[1] - len(prefix) == 0:
                            prefix = ''
                    yield PythonToken(
                        PythonTokenTypes.ENDMARKER, '',
                        (start_pos[0] + line_offset, 0),
                        prefix
                    )
                    break
            elif typ == PythonTokenTypes.NEWLINE and start_pos[0] >= until_line:
                yield PythonToken(typ, string, start_pos, prefix)
                # Check if the parser is actually in a valid suite state.
                if _suite_or_file_input_is_valid(self._pgen_grammar, stack):
                    start_pos = start_pos[0] + 1, 0
                    while len(indents) > int(omitted_first_indent):
                        indents.pop()
                        yield PythonToken(PythonTokenTypes.DEDENT, '', start_pos, '')

                    yield PythonToken(PythonTokenTypes.ENDMARKER, '', start_pos, '')
                    break
                else:
                    continue

            yield PythonToken(typ, string, start_pos, prefix)


class _NodesTreeNode(object):
    _ChildrenGroup = namedtuple('_ChildrenGroup', 'prefix children line_offset last_line_offset_leaf')

    def __init__(self, tree_node, parent=None):
        self.tree_node = tree_node
        self._children_groups = []
        self.parent = parent
        self._node_children = []

    def finish(self):
        children = []
        for prefix, children_part, line_offset, last_line_offset_leaf in self._children_groups:
            first_leaf = _get_next_leaf_if_indentation(
                children_part[0].get_first_leaf()
            )

            first_leaf.prefix = prefix + first_leaf.prefix
            if line_offset != 0:
                try:
                    _update_positions(
                        children_part, line_offset, last_line_offset_leaf)
                except _PositionUpdatingFinished:
                    pass
            children += children_part
        self.tree_node.children = children
        # Reset the parents
        for node in children:
            node.parent = self.tree_node

        for node_child in self._node_children:
            node_child.finish()

    def add_child_node(self, child_node):
        self._node_children.append(child_node)

    def add_tree_nodes(self, prefix, children, line_offset=0, last_line_offset_leaf=None):
        if last_line_offset_leaf is None:
            last_line_offset_leaf = children[-1].get_last_leaf()
        group = self._ChildrenGroup(prefix, children, line_offset, last_line_offset_leaf)
        self._children_groups.append(group)

    def get_last_line(self, suffix):
        line = 0
        if self._children_groups:
            children_group = self._children_groups[-1]
            last_leaf = _get_previous_leaf_if_indentation(
                children_group.last_line_offset_leaf
            )

            line = last_leaf.end_pos[0] + children_group.line_offset

            # Newlines end on the next line, which means that they would cover
            # the next line. That line is not fully parsed at this point.
            if _ends_with_newline(last_leaf, suffix):
                line -= 1
        line += len(split_lines(suffix)) - 1

        if suffix and not suffix.endswith('\n') and not suffix.endswith('\r'):
            # This is the end of a file (that doesn't end with a newline).
            line += 1

        if self._node_children:
            return max(line, self._node_children[-1].get_last_line(suffix))
        return line


class _NodesTree(object):
    def __init__(self, module):
        self._base_node = _NodesTreeNode(module)
        self._working_stack = [self._base_node]
        self._module = module
        self._prefix_remainder = ''
        self.prefix = ''

    @property
    def parsed_until_line(self):
        return self._working_stack[-1].get_last_line(self.prefix)

    def _get_insertion_node(self, indentation_node):
        indentation = indentation_node.start_pos[1]

        # find insertion node
        while True:
            node = self._working_stack[-1]
            tree_node = node.tree_node
            if tree_node.type == 'suite':
                # A suite starts with NEWLINE, ...
                node_indentation = tree_node.children[1].start_pos[1]

                if indentation >= node_indentation:  # Not a Dedent
                    # We might be at the most outer layer: modules. We
                    # don't want to depend on the first statement
                    # having the right indentation.
                    return node

            elif tree_node.type == 'file_input':
                return node

            self._working_stack.pop()

    def add_parsed_nodes(self, tree_nodes):
        old_prefix = self.prefix
        tree_nodes = self._remove_endmarker(tree_nodes)
        if not tree_nodes:
            self.prefix = old_prefix + self.prefix
            return

        assert tree_nodes[0].type != 'newline'

        node = self._get_insertion_node(tree_nodes[0])
        assert node.tree_node.type in ('suite', 'file_input')
        node.add_tree_nodes(old_prefix, tree_nodes)
        # tos = Top of stack
        self._update_tos(tree_nodes[-1])

    def _update_tos(self, tree_node):
        if tree_node.type in ('suite', 'file_input'):
            new_tos = _NodesTreeNode(tree_node)
            new_tos.add_tree_nodes('', list(tree_node.children))

            self._working_stack[-1].add_child_node(new_tos)
            self._working_stack.append(new_tos)

            self._update_tos(tree_node.children[-1])
        elif _func_or_class_has_suite(tree_node):
            self._update_tos(tree_node.children[-1])

    def _remove_endmarker(self, tree_nodes):
        """
        Helps cleaning up the tree nodes that get inserted.
        """
        last_leaf = tree_nodes[-1].get_last_leaf()
        is_endmarker = last_leaf.type == 'endmarker'
        self._prefix_remainder = ''
        if is_endmarker:
            separation = max(last_leaf.prefix.rfind('\n'), last_leaf.prefix.rfind('\r'))
            if separation > -1:
                # Remove the whitespace part of the prefix after a newline.
                # That is not relevant if parentheses were opened. Always parse
                # until the end of a line.
                last_leaf.prefix, self._prefix_remainder = \
                    last_leaf.prefix[:separation + 1], last_leaf.prefix[separation + 1:]

        self.prefix = ''

        if is_endmarker:
            self.prefix = last_leaf.prefix

            tree_nodes = tree_nodes[:-1]
        return tree_nodes

    def copy_nodes(self, tree_nodes, until_line, line_offset):
        """
        Copies tree nodes from the old parser tree.

        Returns the number of tree nodes that were copied.
        """
        if tree_nodes[0].type in ('error_leaf', 'error_node'):
            # Avoid copying errors in the beginning. Can lead to a lot of
            # issues.
            return []

        self._get_insertion_node(tree_nodes[0])

        new_nodes, self._working_stack, self.prefix = self._copy_nodes(
            list(self._working_stack),
            tree_nodes,
            until_line,
            line_offset,
            self.prefix,
        )
        return new_nodes

    def _copy_nodes(self, working_stack, nodes, until_line, line_offset, prefix=''):
        new_nodes = []

        new_prefix = ''
        for node in nodes:
            if node.start_pos[0] > until_line:
                break

            if node.type == 'endmarker':
                break

            if node.type == 'error_leaf' and node.token_type in ('DEDENT', 'ERROR_DEDENT'):
                break
            # TODO this check might take a bit of time for large files. We
            # might want to change this to do more intelligent guessing or
            # binary search.
            if _get_last_line(node) > until_line:
                # We can split up functions and classes later.
                if _func_or_class_has_suite(node):
                    new_nodes.append(node)
                break

            new_nodes.append(node)

        if not new_nodes:
            return [], working_stack, prefix

        tos = working_stack[-1]
        last_node = new_nodes[-1]
        had_valid_suite_last = False
        if _func_or_class_has_suite(last_node):
            suite = last_node
            while suite.type != 'suite':
                suite = suite.children[-1]

            suite_tos = _NodesTreeNode(suite)
            # Don't need to pass line_offset here, it's already done by the
            # parent.
            suite_nodes, new_working_stack, new_prefix = self._copy_nodes(
                working_stack + [suite_tos], suite.children, until_line, line_offset
            )
            if len(suite_nodes) < 2:
                # A suite only with newline is not valid.
                new_nodes.pop()
                new_prefix = ''
            else:
                assert new_nodes
                tos.add_child_node(suite_tos)
                working_stack = new_working_stack
                had_valid_suite_last = True

        if new_nodes:
            last_node = new_nodes[-1]
            if (last_node.type in ('error_leaf', 'error_node') or
                    _is_flow_node(new_nodes[-1])):
                # Error leafs/nodes don't have a defined start/end. Error
                # nodes might not end with a newline (e.g. if there's an
                # open `(`). Therefore ignore all of them unless they are
                # succeeded with valid parser state.
                # If we copy flows at the end, they might be continued
                # after the copy limit (in the new parser).
                # In this while loop we try to remove until we find a newline.
                new_prefix = ''
                new_nodes.pop()
                while new_nodes:
                    last_node = new_nodes[-1]
                    if last_node.get_last_leaf().type == 'newline':
                        break
                    new_nodes.pop()

        if new_nodes:
            if not _ends_with_newline(new_nodes[-1].get_last_leaf()) and not had_valid_suite_last:
                p = new_nodes[-1].get_next_leaf().prefix
                # We are not allowed to remove the newline at the end of the
                # line, otherwise it's going to be missing. This happens e.g.
                # if a bracket is around before that moves newlines to
                # prefixes.
                new_prefix = split_lines(p, keepends=True)[0]

            if had_valid_suite_last:
                last = new_nodes[-1]
                if last.type == 'decorated':
                    last = last.children[-1]
                if last.type in ('async_funcdef', 'async_stmt'):
                    last = last.children[-1]
                last_line_offset_leaf = last.children[-2].get_last_leaf()
                assert last_line_offset_leaf == ':'
            else:
                last_line_offset_leaf = new_nodes[-1].get_last_leaf()
            tos.add_tree_nodes(prefix, new_nodes, line_offset, last_line_offset_leaf)
            prefix = new_prefix
            self._prefix_remainder = ''

        return new_nodes, working_stack, prefix

    def close(self):
        self._base_node.finish()

        # Add an endmarker.
        try:
            last_leaf = self._module.get_last_leaf()
        except IndexError:
            end_pos = [1, 0]
        else:
            last_leaf = _skip_dedent_error_leaves(last_leaf)
            end_pos = list(last_leaf.end_pos)
        lines = split_lines(self.prefix)
        assert len(lines) > 0
        if len(lines) == 1:
            end_pos[1] += len(lines[0])
        else:
            end_pos[0] += len(lines) - 1
            end_pos[1] = len(lines[-1])

        endmarker = EndMarker('', tuple(end_pos), self.prefix + self._prefix_remainder)
        endmarker.parent = self._module
        self._module.children.append(endmarker)



from dataclasses import dataclass, field

from .theme_config_parser import create_palette
from .theme import Theme


@dataclass(frozen=True)
class ThemeConfiguration:
    theme: str = "basic"

    color_scheme: Theme = field(init=False)

    def __post_init__(self):
        # replace theme name with actual palette scheme
        object.__setattr__(self, "color_scheme", create_palette(self.theme))



from typing import Any, List

class reference:
    def __init__(self,
                 rawsource: str = ...,
                 text: str = ...,
                 *children: List[Any],
                 **attributes) -> None: ...

def __getattr__(name) -> Any: ...



from typing import Any, IO, Sequence, Tuple, Union, List, Dict, Protocol, Optional

DEFAULTSECT: str
MAX_INTERPOLATION_DEPTH: int

class Error(Exception):
    message: Any
    def __init__(self, msg: str = ...) -> None: ...
    def _get_message(self) -> None: ...
    def _set_message(self, value: str) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...

class NoSectionError(Error):
    section: str
    def __init__(self, section: str) -> None: ...

class DuplicateSectionError(Error):
    section: str
    def __init__(self, section: str) -> None: ...

class NoOptionError(Error):
    section: str
    option: str
    def __init__(self, option: str, section: str) -> None: ...

class InterpolationError(Error):
    section: str
    option: str
    msg: str
    def __init__(self, option: str, section: str, msg: str) -> None: ...

class InterpolationMissingOptionError(InterpolationError):
    reference: str
    def __init__(self, option: str, section: str, rawval: str, reference: str) -> None: ...

class InterpolationSyntaxError(InterpolationError): ...

class InterpolationDepthError(InterpolationError):
    def __init__(self, option: str, section: str, rawval: str) -> None: ...

class ParsingError(Error):
    filename: str
    errors: List[Tuple[Any, Any]]
    def __init__(self, filename: str) -> None: ...
    def append(self, lineno: Any, line: Any) -> None: ...

class MissingSectionHeaderError(ParsingError):
    lineno: Any
    line: Any
    def __init__(self, filename: str, lineno: Any, line: Any) -> None: ...

class _Readable(Protocol):
    def readline(self) -> str: ...

class RawConfigParser:
    _dict: Any
    _sections: dict
    _defaults: dict
    _optcre: Any
    SECTCRE: Any
    OPTCRE: Any
    OPTCRE_NV: Any
    def __init__(self, defaults: Dict[Any, Any] = ..., dict_type: Any = ..., allow_no_value: bool = ...) -> None: ...
    def defaults(self) -> Dict[Any, Any]: ...
    def sections(self) -> List[str]: ...
    def add_section(self, section: str) -> None: ...
    def has_section(self, section: str) -> bool: ...
    def options(self, section: str) -> List[str]: ...
    def read(self, filenames: Union[str, Sequence[str]]) -> List[str]: ...
    def readfp(self, fp: _Readable, filename: str = ...) -> None: ...
    def get(self, section: str, option: str) -> str: ...
    def items(self, section: str) -> List[Tuple[Any, Any]]: ...
    def _get(self, section: str, conv: type, option: str) -> Any: ...
    def getint(self, section: str, option: str) -> int: ...
    def getfloat(self, section: str, option: str) -> float: ...
    _boolean_states: Dict[str, bool]
    def getboolean(self, section: str, option: str) -> bool: ...
    def optionxform(self, optionstr: str) -> str: ...
    def has_option(self, section: str, option: str) -> bool: ...
    def set(self, section: str, option: str, value: Any = ...) -> None: ...
    def write(self, fp: IO[str]) -> None: ...
    def remove_option(self, section: str, option: Any) -> bool: ...
    def remove_section(self, section: str) -> bool: ...
    def _read(self, fp: IO[str], fpname: str) -> None: ...

class ConfigParser(RawConfigParser):
    _KEYCRE: Any
    def get(self, section: str, option: str, raw: bool = ..., vars: Optional[dict] = ...) -> Any: ...
    def items(self, section: str, raw: bool = ..., vars: Optional[dict] = ...) -> List[Tuple[str, Any]]: ...
    def _interpolate(self, section: str, option: str, rawval: Any, vars: Any) -> str: ...
    def _interpolation_replace(self, match: Any) -> str: ...

class SafeConfigParser(ConfigParser):
    _interpvar_re: Any
    def _interpolate(self, section: str, option: str, rawval: Any, vars: Any) -> str: ...
    def _interpolate_some(self, option: str, accum: list, rest: str, section: str, map: dict, depth: int) -> None: ...



from typing import Any, Optional, Text

def default_stream_factory(total_content_length, filename, content_type, content_length: Optional[Any] = ...): ...
def parse_form_data(environ, stream_factory: Optional[Any] = ..., charset: Text = ..., errors: Text = ...,
                    max_form_memory_size: Optional[Any] = ..., max_content_length: Optional[Any] = ...,
                    cls: Optional[Any] = ..., silent: bool = ...): ...
def exhaust_stream(f): ...

class FormDataParser:
    stream_factory: Any
    charset: Text
    errors: Text
    max_form_memory_size: Any
    max_content_length: Any
    cls: Any
    silent: Any
    def __init__(self, stream_factory: Optional[Any] = ..., charset: Text = ..., errors: Text = ...,
                 max_form_memory_size: Optional[Any] = ..., max_content_length: Optional[Any] = ..., cls: Optional[Any] = ...,
                 silent: bool = ...): ...
    def get_parse_func(self, mimetype, options): ...
    def parse_from_environ(self, environ): ...
    def parse(self, stream, mimetype, content_length, options: Optional[Any] = ...): ...
    parse_functions: Any

def is_valid_multipart_boundary(boundary): ...
def parse_multipart_headers(iterable): ...

class MultiPartParser:
    charset: Text
    errors: Text
    max_form_memory_size: Any
    stream_factory: Any
    cls: Any
    buffer_size: Any
    def __init__(self, stream_factory: Optional[Any] = ..., charset: Text = ..., errors: Text = ...,
                 max_form_memory_size: Optional[Any] = ..., cls: Optional[Any] = ..., buffer_size=...): ...
    def fail(self, message): ...
    def get_part_encoding(self, headers): ...
    def get_part_charset(self, headers) -> Text: ...
    def start_file_streaming(self, filename, headers, total_content_length): ...
    def in_memory_threshold_reached(self, bytes): ...
    def validate_boundary(self, boundary): ...
    def parse_lines(self, file, boundary, content_length, cap_at_buffer: bool = ...): ...
    def parse_parts(self, file, boundary, content_length): ...
    def parse(self, file, boundary, content_length): ...



from typing import Any

from itertools import ifilter as filter
from itertools import imap as map
from itertools import izip as zip


def ascii(obj: Any) -> str: ...


def hex(x: int) -> str: ...


def oct(x: int) -> str: ...



"""
Black Macchiato wrapper that prints errors as stdout and avoids directory permission errors
Lean and mean - faster than importing all of rp
"""
import macchiato
import sys
import io
import os
import tempfile

def autoformat_python_via_black_macchiato(python_code_snippet: str, max_line_length=None) -> str:

    black_args=[]
    if max_line_length is not None:
        assert isinstance(max_line_length, int)
        black_args += ["--line-length", str(max_line_length)]

    input_file = io.StringIO(python_code_snippet)
    output_file = io.StringIO()
    
    exit_code = macchiato.macchiato(input_file, output_file, args=black_args)
    
    if exit_code != 0:
        raise ValueError("rp.autoformat_python_via_black_macchiato: Formatting failed with exit code:", exit_code)
    
    return output_file.getvalue()

try:
    os.chdir(tempfile.gettempdir()) #Make sure we don't get any read-only errors

    code = sys.stdin.read()

    # Format the code
    formatted_code = autoformat_python_via_black_macchiato(code, max_line_length=150)

    if formatted_code.endswith('\n'):
        #Black macchiato adds a new line
        formatted_code = formatted_code[:-1]

    # Print the formatted code
    print(end=formatted_code)

except Exception:
    import rp
    rp.disable_fansi()
    rp.print_verbose_stack_trace()



"""
A collection of JSON rendering related data models and widgets.

Example
=======
.. code-block:: python
   :linenos:

    from rp.libs.pyfx.view.json_lib import JSONListBox, JSONListWalker, NodeFactory, DEFAULT_NODE_IMPLS

    # 1. create NodeFactory
    node_factory = NodeFactory(DEFAULT_NODE_IMPLS)

    # 2. add any custom display implementation
    node_factory.register({
        clazz: json_node_impl
    })

    # 3. create top node from the data (only supports dict, list and primitive variable)
    top_node = node_factory.create_root_node(data)

    # 4. create JSONListBox from top node
    listbox = JSONListBox(JSONListWalker(top_node))

    # 5. use listbox in your own TUI

Implementation Details
======================

Exposed Class
-------------

For integrated this class into your own TUI, three classes is the most and the only entry point.

* :class:`.JSONListBox`
    A :class:`urwid.ListBox` compatible class to manage the visible portion and rendering of the
    JSON tree.
* :class:`.JSONListWalker`
    A :class:`urwid.ListWalker` compatible class to manage the traverse of the whole tree and
    store the current focus node.
* :class:`.NodeFactory`
    A factory to create node based on its value type.

Data Modeling
-------------

The JSON data is loaded into memory as a tree and based on the data type it creates

- Non-leaf Nodes (`array`, `object`)

  Each non-leaf node, two types of nodes are implemented to ease navigation and rendering:

  - Start node / Unexpanded node to represent start / unexpanded line
  - End node to represent end line.

- Leaf Nodes (`string`, `integer`, `numeric`, `boolean`, `null`)

  Each leaf node, single node is enough for navigation.
"""
from .json_listbox import JSONListBox
from .json_listwalker import JSONListWalker
from .node_factory import NodeFactory
from .object import ObjectNode
from .array import ArrayNode
from .primitive import StringNode, IntegerNode, NumericNode, BooleanNode, NullNode, GenericNode
from .selector import DefaultImplementationSelector

DEFAULT_NODE_IMPLS = DefaultImplementationSelector()




from overrides import overrides

from .object_end_node import ObjectEndNode
from .object_start_widget import ObjectStartWidget
from .object_unexpanded_widget import ObjectUnexpandedWidget
from ..json_composite_end_node import JSONCompositeEndNode
from ..json_composite_node import JSONCompositeNode


class ObjectNode(JSONCompositeNode):
    """
    implementation of JSON `object` type node
    aside from fields in a JSONNode, it contains the following elements:
    * value: dict
    * children: dict to store correspondent node
    * sorted_children_key_list: internal type to keep track of a sorted key list
                                and thus keep track of the next, previous node
                                of each child
    * sorted_children_key_list_size: size of key
    """

    def __init__(self, key: str, value: dict, node_factory,
                 parent=None, display_key=True):
        super().__init__(key, value, node_factory, parent, display_key)
        self._children = {}
        self._sorted_children_key_list = sorted(value.keys())
        # avoid re-calculation
        self._sorted_children_key_list_size = len(
            self._sorted_children_key_list
        )

    @overrides
    def collapse_all(self):
        for key, child in self._children.items():
            if isinstance(child, (JSONCompositeNode, JSONCompositeEndNode)):
                child.collapse_all()
        self.toggle_expanded()

    @overrides
    def has_children(self):
        return self._sorted_children_key_list_size != 0

    @overrides
    def get_first_child(self):
        if not self.has_children():
            return None
        return self._get_child_node(self._sorted_children_key_list[0])

    @overrides
    def get_last_child(self):
        if not self.has_children():
            return None
        return self._get_child_node(
            self._sorted_children_key_list[
                self._sorted_children_key_list_size - 1
            ]
        )

    @overrides
    def prev_child(self, key):
        index = self._sorted_children_key_list.index(key)
        if index == 0:
            return None
        return self._get_child_node(self._sorted_children_key_list[index - 1])

    @overrides
    def next_child(self, key):
        index = self._sorted_children_key_list.index(key)
        if index == self._sorted_children_key_list_size - 1:
            return None
        return self._get_child_node(self._sorted_children_key_list[index + 1])

    def _get_child_node(self, key):
        if not self.has_children():
            return None
        elif key not in self._children:
            self._children[key] = self._load_child_node(key)
        return self._children[key]

    def _load_child_node(self, key):
        value = self.get_value()[key]
        return self._node_factory.create_node(
            key, value, parent=self, display_key=True
        )

    # ====================================================================== #
    # ui                                                                     #
    # ====================================================================== #

    @overrides
    def load_unexpanded_widget(self):
        return ObjectUnexpandedWidget(self, self.is_display_key())

    @overrides
    def load_start_widget(self):
        return ObjectStartWidget(self, self.is_display_key())

    @overrides
    def load_end_node(self):
        return ObjectEndNode(self)





from pyflann.exceptions import FLANNException
import binary_dataset
import dat_dataset
import npy_dataset
import hdf5_dataset

import os.path
from numpy import float32

dataset_formats = { 
    'bin' : binary_dataset, 
    'dat' : dat_dataset, 
    'npy' : npy_dataset,
    'hdf5' : hdf5_dataset 
}


def load(filename, rows = -1, cols = -1, dtype = float32, **kwargs):
    
    for format in list(dataset_formats.values()):
        if format.check(filename):
            return format.load(filename, rows, cols, dtype, **kwargs)
    raise FLANNException("Error: Unknown dataset format")
    
    
def save(dataset, filename, format = None, **kwargs):    
    try:
        if format is None:
            basename,extension = os.path.splitext(filename)
            format = extension[1:]
        handler = dataset_formats[format]
        handler.save(dataset, filename, **kwargs)
    except Exception as e:
        raise FLANNException(e)


"""
TODO Some parts of this module are still not well documented.
"""

from rp.libs.jedi.evaluate.context import ModuleContext
from rp.libs.jedi.evaluate import compiled
from rp.libs.jedi.evaluate.compiled import mixed
from rp.libs.jedi.evaluate.compiled.access import create_access_path
from rp.libs.jedi.evaluate.base_context import ContextWrapper


def _create(evaluator, obj):
    return compiled.create_from_access_path(
        evaluator, create_access_path(evaluator, obj)
    )


class NamespaceObject(object):
    def __init__(self, dct):
        self.__dict__ = dct


class MixedModuleContext(ContextWrapper):
    type = 'mixed_module'

    def __init__(self, evaluator, tree_module, namespaces, file_io, code_lines):
        module_context = ModuleContext(
            evaluator, tree_module,
            file_io=file_io,
            string_names=('__main__',),
            code_lines=code_lines
        )
        super(MixedModuleContext, self).__init__(module_context)
        self._namespace_objects = [NamespaceObject(n) for n in namespaces]

    def get_filters(self, *args, **kwargs):
        for filter in self._wrapped_context.get_filters(*args, **kwargs):
            yield filter

        for namespace_obj in self._namespace_objects:
            compiled_object = _create(self.evaluator, namespace_obj)
            mixed_object = mixed.MixedObject(
                compiled_object=compiled_object,
                tree_context=self._wrapped_context
            )
            for filter in mixed_object.get_filters(*args, **kwargs):
                yield filter



import re

from antlr4 import InputStream, CommonTokenStream
from antlr4.error.ErrorListener import ErrorListener
from loguru import logger
from overrides import overrides

from ..common.jsonpath import JSONPathLexer, JSONPathParser, JSONPathListener


def autocomplete(current_input, query):
    """
    Use :class:`.JSONPathAutoCompleteListener`
    to parse the query and give auto-completion suggestions.

    :param current_input: the current query input
    :type current_input: str
    :param query: the query callback which can be used to get data.
    :type query: callback
    """
    input_stream = InputStream(current_input)
    lexer = JSONPathLexer(input_stream)
    lexer.removeErrorListeners()

    stream = CommonTokenStream(lexer)
    parser = JSONPathParser(stream)
    parser.removeErrorListeners()

    listener = JSONPathAutoCompleteListener(query)
    parser.addErrorListener(listener)  # listen syntax error
    parser.addParseListener(listener)  # listen incomplete field name

    # start parse
    parser.jsonpath()

    return listener.is_partial_complete, listener.prefix, listener.options


class JSONPathAutoCompleteListener(JSONPathListener, ErrorListener):
    """
    Both an ErrorListener and JSONPathListener.

    Based on the syntax error ANTLR4 returned to find the possible completions.
    """

    IDENTIFIED_TOKENS = frozenset(['.', '[', '?', '..'])

    def __init__(self, query):
        super().__init__()
        self._query = query
        self._options = list()
        self._prefix = ""
        # flag indicate whether autocomplete only partially complete the query
        self._partial_complete = False
        self._recover_methods = {
            (JSONPathParser.DoubleDotExpressionContext, '..'):
                self.complete_double_dot_field_access,
            (JSONPathParser.SingleDotExpressionContext, '.'):
                self.complete_single_dot_field_access,
            (JSONPathParser.SingleDotExpressionContext, '['):
                self.complete_bracket_field_access,
            (JSONPathParser.FiltersContext, '?'): self.complete_filters,
            (JSONPathParser.UnionContext, '['): self.complete_union
        }

    def reset(self):
        """ reset the current result, #exitFieldAccessor or else may lead to in
        correct conclusion
        """
        self._options = []
        self._prefix = ""
        self._partial_complete = False

    @property
    def options(self):
        return self._options.copy()

    @property
    def prefix(self):
        return self._prefix

    @property
    def is_partial_complete(self):
        return self._partial_complete

    @overrides(check_signature=False)
    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        tokens = recognizer.getTokenStream().tokens
        if len(tokens) <= 1:
            # last token is always a EOF token
            self._options.append('$')
            return
        elif offending_symbol != tokens[-1]:
            # invalid query if offending symbol is not the last token
            self._options.clear()
            return

        self.reset()
        last_token_index = len(tokens) - 2
        identified_token = tokens[last_token_index].text
        while last_token_index >= 0 and \
                identified_token not in self.IDENTIFIED_TOKENS:
            last_token_index -= 1
            identified_token = tokens[last_token_index].text

        key = (type(recognizer._ctx), identified_token)
        try:
            # noinspection PyArgumentList
            self._recover_methods[key](tokens)
        except KeyError:
            logger.opt(exception=True).warning(
                f"{key} not defined in JSONPathAutoCompleteListener.recover_"
                f"methods"
            )
            self._options.clear()

    @overrides
    def exitFieldAccessor(self, ctx: JSONPathParser.FieldAccessorContext):
        self.reset()

        params = dict()
        tokens = ctx.parser.getTokenStream().tokens
        if tokens[-2].text == ']':
            # bypass bracket field, since it's always complete
            self._options = [".", "["]
            self._prefix = ""
            self._partial_complete = True
            return

        last_valid_query = self.find_last_valid_query(tokens)
        current_parent = self._query(last_valid_query)

        if tokens[-4].text == '[*]':
            # this only happens when current parent is list
            current_parent = current_parent[0]
            params["include_wildcard"] = False
        params["parent"] = current_parent
        params["prefix"] = tokens[-2].text

        options = self.find_options(**params)

        if len(options) == 1 and options[0] == tokens[-2].text:
            # the only current options is complete, suggest the next token
            self._options = [".", "["]
            self._prefix = ""
            self._partial_complete = True
            return

        self._options = options
        self._prefix = tokens[-2].text

    def complete_double_dot_field_access(self, tokens):
        pass

    def complete_single_dot_field_access(self, tokens):
        params = dict()

        last_valid_query = self.find_last_valid_query(
            tokens, optional_single_dot=False
        )
        current_parent = self._query(last_valid_query)

        if tokens[-3].text == '[*]':
            # this only happens when current parent is list
            current_parent = current_parent[0]
            params["include_wildcard"] = False

        params["parent"] = current_parent
        self._options = self.find_options(**params)
        self._prefix = ""

    def complete_bracket_field_access(self, tokens):
        params = dict()
        # bracket field or array index
        if tokens[-2].text != '[':
            last_valid_query_end = -3
        else:
            last_valid_query_end = -2
        last_valid_query = self.find_last_valid_query(
            tokens, last_valid_query_end=last_valid_query_end
        )
        current_parent = self._query(last_valid_query)

        if tokens[last_valid_query_end - 1].text == '[*]' or \
                (tokens[last_valid_query_end - 1].text == '.' and
                 tokens[last_valid_query_end - 2].text == '[*]'):
            # this only happens when current parent is list
            current_parent = current_parent[0]
            params["include_wildcard"] = False

        params["parent"] = current_parent
        params["prefix"] = ''.join(
            [t.text for t in tokens[last_valid_query_end:-1]]
        )
        self._options = self.find_options(**params)
        self._prefix = params["prefix"]

    def complete_filters(self, tokens):
        params = {
            "include_wildcard": False
        }
        contains_braces = False
        question_mark_index = len(tokens) - 2
        while question_mark_index >= 0 and \
                tokens[question_mark_index].text != '?':
            contains_braces = tokens[question_mark_index].text == '('
            question_mark_index -= 1

        if not contains_braces:
            self._options = ['(']
            self._prefix = ""
            self._partial_complete = True
            return

        last_valid_query = self.find_last_valid_query(
            tokens, last_valid_query_end=question_mark_index - 1
        )
        current_parent = self._query(last_valid_query)

        if tokens[question_mark_index - 2].text == '[*]' or \
                (tokens[question_mark_index - 2].text == '.' and
                 tokens[question_mark_index - 3].text == '[*]'):
            # this only happens when current parent is list
            current_parent = current_parent[0]
            params["include_wildcard"] = False

        params["parent"] = current_parent
        self._options = [f"@.{o}" for o in self.find_options(**params)]
        self._prefix = ""
        self._partial_complete = True

    def complete_union(self, tokens):
        params = {
            "include_wildcard": False,
            "is_union": True
        }
        existed_keys = set()
        last_valid_query_end = len(tokens) - 2
        while last_valid_query_end >= 0 and \
                tokens[last_valid_query_end].text != '[':
            last_valid_query_end -= 1
            if re.match(r"^'.*'$", tokens[last_valid_query_end].text):
                existed_keys.add(tokens[last_valid_query_end].text)
        last_valid_query = self.find_last_valid_query(
            tokens, last_valid_query_end=last_valid_query_end
        )
        current_parent = self._query(last_valid_query)

        if tokens[last_valid_query_end -
                  1].text == '[*]' or (tokens[last_valid_query_end -
                                              1].text == '.' and tokens[
                      last_valid_query_end -
                      2].text == '[*]'):
            # this only happens when current parent is list
            current_parent = current_parent[0]
            params["include_wildcard"] = False

        params["parent"] = current_parent
        params["prefix"] = tokens[-2].text if tokens[-2].text != ',' else ""
        options = self.find_options(**params)
        self._options = list(filter(lambda o: o not in existed_keys, options))
        self._prefix = params["prefix"]
        self._partial_complete = True

    @staticmethod
    def find_last_valid_query(
            tokens, last_valid_query_end=-2, optional_single_dot=True):
        # last token will always be EOF
        if optional_single_dot and tokens[last_valid_query_end - 1].text == '.':
            return ''.join([t.text for t in tokens[:last_valid_query_end - 1]])
        return ''.join([t.text for t in tokens[:last_valid_query_end]])

    @staticmethod
    def find_options(parent, prefix="", include_wildcard=True, is_union=False):
        options = []
        if isinstance(parent, list):
            # current parent is a list
            if include_wildcard and not is_union:
                options.append('[*]')
            options.extend(
                JSONPathAutoCompleteListener._generate_list_completes(
                    len(parent), prefix, is_union
                )
            )
        elif isinstance(parent, dict):
            if include_wildcard and not is_union:
                if prefix.startswith('['):
                    options.append('[*]')
                else:
                    options.append('*')
            for key in parent.keys():
                if is_union:
                    options.append(f"'{key}'")
                elif (not re.match(r'\w+', key)) or prefix.startswith('['):
                    options.append("['" + key + "']")
                else:
                    options.append(key)
            options = list(filter(lambda k: k.startswith(prefix), options))
        return options

    @staticmethod
    def _generate_list_completes(length, prefix="", is_union=False):
        """
        Generates completes for an array, use its length also add `*` to represent
        everything.

        :param length: the list length
        :type length: int
        :param prefix: the current prefix
        :type prefix: str
        :param is_union: whether the complete method is a union method
        :type is_union: bool
        :return: completes for current length.
        """
        if is_union:
            options = [str(index) for index in range(length)]
        else:
            options = ['[' + str(index) + ']' for index in range(length)]
        return list(filter(lambda k: k.startswith(prefix), options))



import sys

from typing import Any
from base64 import encodestring as encodebytes

from six.moves import http_client

expanduser: Any

if sys.version_info >= (3, 0):
    StandardError = Exception
else:
    from __builtin__ import StandardError as StandardError

long_type: Any
unquote_str: Any
parse_qs_safe: Any



import sys
from typing import Any, Optional

if sys.version_info < (3,):
    from SocketServer import ThreadingMixIn, ForkingMixIn
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
else:
    from socketserver import ThreadingMixIn, ForkingMixIn
    from http.server import HTTPServer, BaseHTTPRequestHandler

class _SslDummy:
    def __getattr__(self, name): ...

ssl: Any
LISTEN_QUEUE: Any
can_open_by_fd: Any

class WSGIRequestHandler(BaseHTTPRequestHandler):
    @property
    def server_version(self): ...
    def make_environ(self): ...
    environ: Any
    close_connection: Any
    def run_wsgi(self): ...
    def handle(self): ...
    def initiate_shutdown(self): ...
    def connection_dropped(self, error, environ: Optional[Any] = ...): ...
    raw_requestline: Any
    def handle_one_request(self): ...
    def send_response(self, code, message: Optional[Any] = ...): ...
    def version_string(self): ...
    def address_string(self): ...
    def port_integer(self): ...
    def log_request(self, code: object = ..., size: object = ...) -> None: ...
    def log_error(self, *args): ...
    def log_message(self, format, *args): ...
    def log(self, type, message, *args): ...

BaseRequestHandler: Any

def generate_adhoc_ssl_pair(cn: Optional[Any] = ...): ...
def make_ssl_devcert(base_path, host: Optional[Any] = ..., cn: Optional[Any] = ...): ...
def generate_adhoc_ssl_context(): ...
def load_ssl_context(cert_file, pkey_file: Optional[Any] = ..., protocol: Optional[Any] = ...): ...

class _SSLContext:
    def __init__(self, protocol): ...
    def load_cert_chain(self, certfile, keyfile: Optional[Any] = ..., password: Optional[Any] = ...): ...
    def wrap_socket(self, sock, **kwargs): ...

def is_ssl_error(error: Optional[Any] = ...): ...
def select_ip_version(host, port): ...

class BaseWSGIServer(HTTPServer):
    multithread: Any
    multiprocess: Any
    request_queue_size: Any
    address_family: Any
    app: Any
    passthrough_errors: Any
    shutdown_signal: Any
    host: Any
    port: Any
    socket: Any
    server_address: Any
    ssl_context: Any
    def __init__(self, host, port, app, handler: Optional[Any] = ..., passthrough_errors: bool = ...,
                 ssl_context: Optional[Any] = ..., fd: Optional[Any] = ...): ...
    def log(self, type, message, *args): ...
    def serve_forever(self): ...
    def handle_error(self, request, client_address): ...
    def get_request(self): ...

class ThreadedWSGIServer(ThreadingMixIn, BaseWSGIServer):
    multithread: Any
    daemon_threads: Any

class ForkingWSGIServer(ForkingMixIn, BaseWSGIServer):
    multiprocess: Any
    max_children: Any
    def __init__(self, host, port, app, processes: int = ..., handler: Optional[Any] = ..., passthrough_errors: bool = ...,
                 ssl_context: Optional[Any] = ..., fd: Optional[Any] = ...): ...

def make_server(host: Optional[Any] = ..., port: Optional[Any] = ..., app: Optional[Any] = ..., threaded: bool = ...,
                processes: int = ..., request_handler: Optional[Any] = ..., passthrough_errors: bool = ...,
                ssl_context: Optional[Any] = ..., fd: Optional[Any] = ...): ...
def is_running_from_reloader(): ...
def run_simple(hostname, port, application, use_reloader: bool = ..., use_debugger: bool = ..., use_evalex: bool = ...,
               extra_files: Optional[Any] = ..., reloader_interval: int = ..., reloader_type: str = ..., threaded: bool = ...,
               processes: int = ..., request_handler: Optional[Any] = ..., static_files: Optional[Any] = ...,
               passthrough_errors: bool = ..., ssl_context: Optional[Any] = ...): ...
def run_with_reloader(*args, **kwargs): ...
def main(): ...



from dataclasses import dataclass

from overrides import overrides

from .autocomplete_popup import AutoCompletePopUpKeys
from ...keymapper import AbstractComponentKeyMapper


@dataclass(frozen=True)
class AutoCompletePopUpKeyMapper(AbstractComponentKeyMapper):

    cursor_up: str = "up"
    cursor_down: str = "down"
    select: str = "enter"
    cancel: str = "esc"

    @property
    @overrides
    def mapped_key(self):
        return {
            self.cursor_up: AutoCompletePopUpKeys.CURSOR_UP,
            self.cursor_down: AutoCompletePopUpKeys.CURSOR_DOWN,
            self.select: AutoCompletePopUpKeys.SELECT,
            self.cancel: AutoCompletePopUpKeys.CANCEL
        }



"""
"""
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod
from six import with_metaclass

__all__ = (
    'Completion',
    'Completer',
    'CompleteEvent',
    'get_common_complete_suffix',
)


class Completion(object):
    """
    :param text: The new string that will be inserted into the document.
    :param start_position: Position relative to the cursor_position where the
        new text will start. The text will be inserted between the
        start_position and the original cursor position.
    :param display: (optional string) If the completion has to be displayed
        differently in the completion menu.
    :param display_meta: (Optional string) Meta information about the
        completion, e.g. the path or source where it's coming from.
    :param get_display_meta: Lazy `display_meta`. Retrieve meta information
        only when meta is displayed.
    """
    def __init__(self, text, start_position=0, display=None, display_meta=None,
                 get_display_meta=None):
        self.text = text
        self.start_position = start_position
        self._display_meta = display_meta
        self._get_display_meta = get_display_meta

        if display is None:
            self.display = text
        else:
            self.display = display

        assert self.start_position <= 0

    def __repr__(self):
        if self.display == self.text:
            return '%s(text=%r, start_position=%r)' % (
                self.__class__.__name__, self.text, self.start_position)
        else:
            return '%s(text=%r, start_position=%r, display=%r)' % (
                self.__class__.__name__, self.text, self.start_position,
                self.display)

    def __eq__(self, other):
        return (
            self.text == other.text and
            self.start_position == other.start_position and
            self.display == other.display and
            self.display_meta == other.display_meta)

    def __hash__(self):
        return hash((self.text, self.start_position, self.display, self.display_meta))

    @property
    def display_meta(self):
        # Return meta-text. (This is lazy when using "get_display_meta".)
        if self._display_meta is not None:
            return self._display_meta

        elif self._get_display_meta:
            self._display_meta = self._get_display_meta()
            return self._display_meta

        else:
            return ''

    def new_completion_from_position(self, position):
        """
        (Only for internal use!)
        Get a new completion by splitting this one. Used by
        `CommandLineInterface` when it needs to have a list of new completions
        after inserting the common prefix.
        """
        assert isinstance(position, int) and position - self.start_position >= 0

        return Completion(
            text=self.text[position - self.start_position:],
            display=self.display,
            display_meta=self._display_meta,
            get_display_meta=self._get_display_meta)


class CompleteEvent(object):
    """
    Event that called the completer.

    :param text_inserted: When True, it means that completions are requested
        because of a text insert. (`Buffer.complete_while_typing`.)
    :param completion_requested: When True, it means that the user explicitely
        pressed the `Tab` key in order to view the completions.

    These two flags can be used for instance to implemented a completer that
    shows some completions when ``Tab`` has been pressed, but not
    automatically when the user presses a space. (Because of
    `complete_while_typing`.)
    """
    def __init__(self, text_inserted=False, completion_requested=False):
        assert not (text_inserted and completion_requested)

        #: Automatic completion while typing.
        self.text_inserted = text_inserted

        #: Used explicitely requested completion by pressing 'tab'.
        self.completion_requested = completion_requested

    def __repr__(self):
        return '%s(text_inserted=%r, completion_requested=%r)' % (
            self.__class__.__name__, self.text_inserted, self.completion_requested)


class Completer(with_metaclass(ABCMeta, object)):
    """
    Base class for completer implementations.
    """
    @abstractmethod
    def get_completions(self, document, complete_event):
        """
        Yield :class:`.Completion` instances.

        :param document: :class:`~prompt_toolkit.document.Document` instance.
        :param complete_event: :class:`.CompleteEvent` instance.
        """
        while False:
            yield


def get_common_complete_suffix(document, completions):
    """
    Return the common prefix for all completions.
    """
    # Take only completions that don't change the text before the cursor.
    def doesnt_change_before_cursor(completion):
        end = completion.text[:-completion.start_position]
        return document.text_before_cursor.endswith(end)

    completions2 = [c for c in completions if doesnt_change_before_cursor(c)]

    # When there is at least one completion that changes the text before the
    # cursor, don't return any common part.
    if len(completions2) != len(completions):
        return ''

    # Return the common prefix.
    def get_suffix(completion):
        return completion.text[-completion.start_position:]

    return _commonprefix([get_suffix(c) for c in completions2])


def _commonprefix(strings):
    # Similar to os.path.commonprefix
    if not strings:
        return ''

    else:
        s1 = min(strings)
        s2 = max(strings)

        for i, c in enumerate(s1):
            if c != s2[i]:
                return s1[:i]

        return s1



"""
Helpers for the API
"""
import re
from collections import namedtuple
from textwrap import dedent

from rp.libs.parso.python.parser import Parser
from rp.libs.parso.python import tree

from rp.libs.jedi._compatibility import u, Parameter
from rp.libs.jedi.evaluate.base_context import NO_CONTEXTS
from rp.libs.jedi.evaluate.syntax_tree import eval_atom
from rp.libs.jedi.evaluate.helpers import evaluate_call_of_leaf
from rp.libs.jedi.evaluate.compiled import get_string_context_set
from rp.libs.jedi.cache import call_signature_time_cache


CompletionParts = namedtuple('CompletionParts', ['path', 'has_dot', 'name'])


def sorted_definitions(defs):
    # Note: `or ''` below is required because `module_path` could be
    return sorted(defs, key=lambda x: (x.module_path or '', x.line or 0, x.column or 0, x.name))


def get_on_completion_name(module_node, lines, position):
    leaf = module_node.get_leaf_for_position(position)
    if leaf is None or leaf.type in ('string', 'error_leaf'):
        # Completions inside strings are a bit special, we need to parse the
        # string. The same is true for comments and error_leafs.
        line = lines[position[0] - 1]
        # The first step of completions is to get the name
        return re.search(r'(?!\d)\w+$|$', line[:position[1]]).group(0)
    elif leaf.type not in ('name', 'keyword'):
        return ''

    return leaf.value[:position[1] - leaf.start_pos[1]]


def _get_code(code_lines, start_pos, end_pos):
    # Get relevant lines.
    lines = code_lines[start_pos[0] - 1:end_pos[0]]
    # Remove the parts at the end of the line.
    lines[-1] = lines[-1][:end_pos[1]]
    # Remove first line indentation.
    lines[0] = lines[0][start_pos[1]:]
    return ''.join(lines)


class OnErrorLeaf(Exception):
    @property
    def error_leaf(self):
        return self.args[0]


def _get_code_for_stack(code_lines, leaf, position):
    # It might happen that we're on whitespace or on a comment. This means
    # that we would not get the right leaf.
    if leaf.start_pos >= position:
        # If we're not on a comment simply get the previous leaf and proceed.
        leaf = leaf.get_previous_leaf()
        if leaf is None:
            return u('')  # At the beginning of the file.

    is_after_newline = leaf.type == 'newline'
    while leaf.type == 'newline':
        leaf = leaf.get_previous_leaf()
        if leaf is None:
            return u('')

    if leaf.type == 'error_leaf' or leaf.type == 'string':
        if leaf.start_pos[0] < position[0]:
            # On a different line, we just begin anew.
            return u('')

        # Error leafs cannot be parsed, completion in strings is also
        # impossible.
        raise OnErrorLeaf(leaf)
    else:
        user_stmt = leaf
        while True:
            if user_stmt.parent.type in ('file_input', 'suite', 'simple_stmt'):
                break
            user_stmt = user_stmt.parent

        if is_after_newline:
            if user_stmt.start_pos[1] > position[1]:
                # This means that it's actually a dedent and that means that we
                # start without context (part of a suite).
                return u('')

        # This is basically getting the relevant lines.
        return _get_code(code_lines, user_stmt.get_start_pos_of_prefix(), position)


def get_stack_at_position(grammar, code_lines, leaf, pos):
    """
    Returns the possible node names (e.g. import_from, xor_test or yield_stmt).
    """
    class EndMarkerReached(Exception):
        pass

    def tokenize_without_endmarker(code):
        # TODO This is for now not an official parso API that exists purely
        #   for Jedi.
        tokens = grammar._tokenize(code)
        for token in tokens:
            if token.string == safeword:
                raise EndMarkerReached()
            elif token.prefix.endswith(safeword):
                # This happens with comments.
                raise EndMarkerReached()
            elif token.string.endswith(safeword):
                yield token  # Probably an f-string literal that was not finished.
                raise EndMarkerReached()
            else:
                yield token

    # The code might be indedented, just remove it.
    code = dedent(_get_code_for_stack(code_lines, leaf, pos))
    # We use a word to tell Jedi when we have reached the start of the
    # completion.
    # Use Z as a prefix because it's not part of a number suffix.
    safeword = 'ZZZ_USER_WANTS_TO_COMPLETE_HERE_WITH_JEDI'
    code = code + ' ' + safeword

    p = Parser(grammar._pgen_grammar, error_recovery=True)
    try:
        p.parse(tokens=tokenize_without_endmarker(code))
    except EndMarkerReached:
        return p.stack
    raise SystemError(
        "This really shouldn't happen. There's a bug in Jedi:\n%s"
        % list(tokenize_without_endmarker(code))
    )


def evaluate_goto_definition(evaluator, context, leaf):
    if leaf.type == 'name':
        # In case of a name we can just use goto_definition which does all the
        # magic itself.
        return evaluator.goto_definitions(context, leaf)

    parent = leaf.parent
    definitions = NO_CONTEXTS
    if parent.type == 'atom':
        # e.g. `(a + b)`
        definitions = context.eval_node(leaf.parent)
    elif parent.type == 'trailer':
        # e.g. `a()`
        definitions = evaluate_call_of_leaf(context, leaf)
    elif isinstance(leaf, tree.Literal):
        # e.g. `"foo"` or `1.0`
        return eval_atom(context, leaf)
    elif leaf.type in ('fstring_string', 'fstring_start', 'fstring_end'):
        return get_string_context_set(evaluator)
    return definitions


class CallDetails(object):
    def __init__(self, bracket_leaf, children, position):
        ['bracket_leaf', 'call_index', 'keyword_name_str']
        self.bracket_leaf = bracket_leaf
        self._children = children
        self._position = position

    @property
    def index(self):
        return _get_index_and_key(self._children, self._position)[0]

    @property
    def keyword_name_str(self):
        return _get_index_and_key(self._children, self._position)[1]

    def calculate_index(self, param_names):
        positional_count = 0
        used_names = set()
        star_count = -1
        args = list(_iter_arguments(self._children, self._position))
        if not args:
            if param_names:
                return 0
            else:
                return None

        is_kwarg = False
        for i, (star_count, key_start, had_equal) in enumerate(args):
            is_kwarg |= had_equal | (star_count == 2)
            if star_count:
                pass  # For now do nothing, we don't know what's in there here.
            else:
                if i + 1 != len(args):  # Not last
                    if had_equal:
                        used_names.add(key_start)
                    else:
                        positional_count += 1

        for i, param_name in enumerate(param_names):
            kind = param_name.get_kind()

            if not is_kwarg:
                if kind == Parameter.VAR_POSITIONAL:
                    return i
                if kind in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.POSITIONAL_ONLY):
                    if i == positional_count:
                        return i

            if key_start is not None and not star_count == 1 or star_count == 2:
                if param_name.string_name not in used_names \
                        and (kind == Parameter.KEYWORD_ONLY
                             or kind == Parameter.POSITIONAL_OR_KEYWORD
                             and positional_count <= i):
                    if star_count:
                        return i
                    if had_equal:
                        if param_name.string_name == key_start:
                            return i
                    else:
                        if param_name.string_name.startswith(key_start):
                            return i

                if kind == Parameter.VAR_KEYWORD:
                    return i
        return None


def _iter_arguments(nodes, position):
    def remove_after_pos(name):
        if name.type != 'name':
            return None
        return name.value[:position[1] - name.start_pos[1]]

    # Returns Generator[Tuple[star_count, Optional[key_start: str], had_equal]]
    nodes_before = [c for c in nodes if c.start_pos < position]
    if nodes_before[-1].type == 'arglist':
        for x in _iter_arguments(nodes_before[-1].children, position):
            yield x  # Python 2 :(
        return

    previous_node_yielded = False
    stars_seen = 0
    for i, node in enumerate(nodes_before):
        if node.type == 'argument':
            previous_node_yielded = True
            first = node.children[0]
            second = node.children[1]
            if second == '=':
                if second.start_pos < position:
                    yield 0, first.value, True
                else:
                    yield 0, remove_after_pos(first), False
            elif first in ('*', '**'):
                yield len(first.value), remove_after_pos(second), False
            else:
                # Must be a Comprehension
                first_leaf = node.get_first_leaf()
                if first_leaf.type == 'name' and first_leaf.start_pos >= position:
                    yield 0, remove_after_pos(first_leaf), False
                else:
                    yield 0, None, False
            stars_seen = 0
        elif node.type in ('testlist', 'testlist_star_expr'):  # testlist is Python 2
            for n in node.children[::2]:
                if n.type == 'star_expr':
                    stars_seen = 1
                    n = n.children[1]
                yield stars_seen, remove_after_pos(n), False
                stars_seen = 0
            # The count of children is even if there's a comma at the end.
            previous_node_yielded = bool(len(node.children) % 2)
        elif isinstance(node, tree.PythonLeaf) and node.value == ',':
            if not previous_node_yielded:
                yield stars_seen, '', False
                stars_seen = 0
            previous_node_yielded = False
        elif isinstance(node, tree.PythonLeaf) and node.value in ('*', '**'):
            stars_seen = len(node.value)
        elif node == '=' and nodes_before[-1]:
            previous_node_yielded = True
            before = nodes_before[i - 1]
            if before.type == 'name':
                yield 0, before.value, True
            else:
                yield 0, None, False
            # Just ignore the star that is probably a syntax error.
            stars_seen = 0

    if not previous_node_yielded:
        if nodes_before[-1].type == 'name':
            yield stars_seen, remove_after_pos(nodes_before[-1]), False
        else:
            yield stars_seen, '', False


def _get_index_and_key(nodes, position):
    """
    Returns the amount of commas and the keyword argument string.
    """
    nodes_before = [c for c in nodes if c.start_pos < position]
    if nodes_before[-1].type == 'arglist':
        return _get_index_and_key(nodes_before[-1].children, position)

    key_str = None

    last = nodes_before[-1]
    if last.type == 'argument' and last.children[1] == '=' \
            and last.children[1].end_pos <= position:
        # Checked if the argument
        key_str = last.children[0].value
    elif last == '=':
        key_str = nodes_before[-2].value

    return nodes_before.count(','), key_str


def _get_call_signature_details_from_error_node(node, additional_children, position):
    for index, element in reversed(list(enumerate(node.children))):
        # `index > 0` means that it's a trailer and not an atom.
        if element == '(' and element.end_pos <= position and index > 0:
            # It's an error node, we don't want to match too much, just
            # until the parentheses is enough.
            children = node.children[index:]
            name = element.get_previous_leaf()
            if name is None:
                continue
            if name.type == 'name' or name.parent.type in ('trailer', 'atom'):
                return CallDetails(element, children + additional_children, position)


def get_call_signature_details(module, position):
    leaf = module.get_leaf_for_position(position, include_prefixes=True)
    if leaf.start_pos >= position:
        # Whitespace / comments after the leaf count towards the previous leaf.
        leaf = leaf.get_previous_leaf()
        if leaf is None:
            return None

    if leaf == ')':
        # TODO is this ok?
        if leaf.end_pos == position:
            leaf = leaf.get_next_leaf()

    # Now that we know where we are in the syntax tree, we start to look at
    # parents for possible function definitions.
    node = leaf.parent
    while node is not None:
        if node.type in ('funcdef', 'classdef'):
            # Don't show call signatures if there's stuff before it that just
            # makes it feel strange to have a call signature.
            return None

        additional_children = []
        for n in reversed(node.children):
            if n.start_pos < position:
                if n.type == 'error_node':
                    result = _get_call_signature_details_from_error_node(
                        n, additional_children, position
                    )
                    if result is not None:
                        return result

                    additional_children[0:0] = n.children
                    continue
                additional_children.insert(0, n)

        if node.type == 'trailer' and node.children[0] == '(':
            leaf = node.get_previous_leaf()
            if leaf is None:
                return None
            return CallDetails(node.children[0], node.children, position)

        node = node.parent

    return None


@call_signature_time_cache("call_signatures_validity")
def cache_call_signatures(evaluator, context, bracket_leaf, code_lines, user_pos):
    """This function calculates the cache key."""
    line_index = user_pos[0] - 1

    before_cursor = code_lines[line_index][:user_pos[1]]
    other_lines = code_lines[bracket_leaf.start_pos[0]:line_index]
    whole = ''.join(other_lines + [before_cursor])
    before_bracket = re.match(r'.*\(', whole, re.DOTALL)

    module_path = context.get_root_context().py__file__()
    if module_path is None:
        yield None  # Don't cache!
    else:
        yield (module_path, before_bracket, bracket_leaf.start_pos)
    yield evaluate_goto_definition(
        evaluator,
        context,
        bracket_leaf.get_previous_leaf(),
    )



"""
Example
=======
.. code-block:: python
   :linenos:

   from rp.libs.pyfx import Controller

   # data is the what you want to render as TUI
   # only supports dict, list and primitive variable
   Controller().run_with_data(data)
"""
from .config import Configuration
from .model import Model
from .service.client import Client
from .service.dispatcher import Dispatcher
from .view import View


class PyfxApp:
    """
    *Pyfx* controller, the main entry point of pyfx library.
    """

    def __init__(self, config=Configuration()):
        self._config = config
        self._dispatcher = Dispatcher()
        self._client = Client(self._dispatcher)
        self._view = View(config.view, self._client)
        self._model = Model(self._dispatcher)

    def run(self, type, *args):
        data = self._model.load(type, *args)
        self._view.run(data)

    def exit(self, exception):
        self._view.exit(exception)
        self._client.shutdown(wait=False)



from abc import ABCMeta
from abc import abstractmethod

from overrides import final
from overrides import overrides

from .json_simple_node import JSONSimpleNode


class JSONCompositeNode(JSONSimpleNode, metaclass=ABCMeta):
    """
    base node represents a JSON `object` or `array` type, also a non-leaf node in the
    whole parsed tree.
    """

    def __init__(self, key, value, node_factory, parent=None, display_key=True
                 ):
        super().__init__(key, value, parent, display_key)
        self._node_factory = node_factory
        # only display the first layer on start
        self._expanded = self.is_root()
        self._start_widget = None
        self._unexpanded_widget = None
        self._end_node = None

    # =================================================================================== #
    # getters and setters                                                                 #
    # =================================================================================== #

    # expanded
    def is_expanded(self):
        return self._expanded

    def toggle_expanded(self):
        self._expanded = not self._expanded

    @abstractmethod
    def collapse_all(self):
        pass

    @abstractmethod
    def has_children(self):
        pass

    @abstractmethod
    def get_first_child(self):
        pass

    @abstractmethod
    def get_last_child(self):
        pass

    @abstractmethod
    def prev_child(self, key):
        pass

    @abstractmethod
    def next_child(self, key):
        pass

    # end_node
    def get_end_node(self):
        if not self.is_expanded():
            # unexpanded node is itself the end node
            return self
        if self._end_node is None:
            self._end_node = self.load_end_node()
        return self._end_node

    @abstractmethod
    def load_end_node(self):
        pass

    # =================================================================================== #
    # ui                                                                                  #
    # =================================================================================== #

    @final
    @overrides
    def get_widget(self):
        if not self.is_expanded():
            return self.get_unexpanded_widget()
        return self.get_start_widget()

    @final
    @overrides
    def load_widget(self):
        raise NotImplementedError(
            f"{type(self)} is a composite node and does not have #load_widget() method."
        )

    # start widget
    def get_start_widget(self):
        if self._start_widget is None:
            self._start_widget = self.load_start_widget()
        return self._start_widget

    @abstractmethod
    def load_start_widget(self):
        pass

    # unexpanded widget
    def get_unexpanded_widget(self):
        if self._unexpanded_widget is None:
            self._unexpanded_widget = self.load_unexpanded_widget()
        return self._unexpanded_widget

    @abstractmethod
    def load_unexpanded_widget(self):
        pass
